import pytest

from pyartnet.output_correction import cubic, quadratic, quadruple


@pytest.mark.parametrize('max_val', [
    pytest.param(k, id=f'{k:X}') for k in (0xFF, 0xFFFF, 0xFFFFFF, 0xFFFFFFFF, 0xFFFFFFFFFF)])
@pytest.mark.parametrize('corr', [quadratic, quadruple, cubic])
def test_correction(corr, max_val):
    assert corr(0, max_val=max_val) == 0
    assert corr(max_val, max_val=max_val) == max_val



import logging
import warnings
from array import array
from logging import DEBUG as LVL_DEBUG
from math import ceil
from typing import Any, Callable, Collection, Final, List, Literal, Optional, Type, Union

from pyartnet.errors import ChannelOutOfUniverseError, ChannelValueOutOfBoundsError, \
    ChannelWidthError, ValueCountDoesNotMatchChannelWidthError
from pyartnet.output_correction import linear

from ..fades import FadeBase, LinearFade
from .channel_fade import ChannelBoundFade
from .output_correction import OutputCorrection
from .universe import BaseUniverse

log = logging.getLogger('pyartnet.Channel')


ARRAY_TYPE: Final = {
    1: 'B',  # unsigned char : min size 1 byte
    2: 'H',  # unsigned short: min size 2 bytes
    3: 'L',  # unsigned long : min size 4 bytes
    4: 'L'   # unsigned long : min size 4 bytes
}


class Channel(OutputCorrection):
    def __init__(self, universe: BaseUniverse,
                 start: int, width: int,
                 byte_size: int = 1, byte_order: Literal['big', 'little'] = 'little'):
        super().__init__()

        # Validate Boundaries
        if byte_size not in ARRAY_TYPE:
            raise ValueError(f'Value size must be {", ".join(map(str, ARRAY_TYPE))}')

        if start < 1 or start > 512:
            raise ChannelOutOfUniverseError(
                f'Start position of channel out of universe (1..512): {start}')

        if width <= 0 or not isinstance(width, int):
            raise ChannelWidthError(
                f'Channel width must be int > 0: {width} ({type(width)})')

        total_byte_width: Final = width * byte_size

        self._start: Final = start
        self._width: Final = width
        self._stop: Final = start + total_byte_width - 1

        if self._stop > 512:
            raise ChannelOutOfUniverseError(
                f'End position of channel out of universe (1..512): '
                f'start: {self._start} width: {self._width} * {byte_size}bytes -> {self._stop}'
            )

        # value representation
        self._byte_size: Final = byte_size
        self._byte_order: Final = byte_order
        self._value_max: Final = 256 ** self._byte_size - 1
        self._buf_start: Final = self._start - 1

        null_vals = [0 for _ in range(self._width)]
        self._values_raw: array[int] = array(ARRAY_TYPE[self._byte_size], null_vals)    # uncorrected values
        self._values_act: array[int] = array(ARRAY_TYPE[self._byte_size], null_vals)    # values after output correction

        # Parents
        self._parent_universe: Final = universe
        self._parent_node: Final = universe._node

        self._correction_current: Callable[[float, int], float] = linear

        # Fade
        self._current_fade: Optional[ChannelBoundFade] = None

        # ---------------------------------------------------------------------
        # Values that can be set by the user
        # ---------------------------------------------------------------------
        # Callbacks
        self.callback_fade_finished: Optional[Callable[[Channel], Any]] = None

    def _apply_output_correction(self):
        # default correction is linear
        self._correction_current = linear

        # inherit correction if it is not set first from universe and then from the node
        for obj in (self, self._parent_universe, self._parent_node):
            if obj._correction_output is not None:
                self._correction_current = obj._correction_output
                return None

    def get_values(self) -> List[int]:
        """Get the current (uncorrected) channel values

        :return: list of channel values
        """
        return self._values_raw.tolist()

    def set_values(self, values: Collection[Union[int, float]]):
        """Set values for a channel without a fade

        :param values: Iterable of values with the same size as the channel width
        """
        # get output correction function
        if len(values) != self._width:
            raise ValueCountDoesNotMatchChannelWidthError(
                f'Not enough fade values specified, expected {self._width} but got {len(values)}!')

        correction = self._correction_current
        value_max = self._value_max

        changed = False
        for i, val in enumerate(values):
            raw_new = round(val)
            if not 0 <= raw_new <= value_max:
                raise ChannelValueOutOfBoundsError(f'Channel value out of bounds! 0 <= {val} <= {value_max:d}')

            self._values_raw[i] = raw_new
            act_new = round(correction(val, value_max)) if correction is not linear else raw_new
            if self._values_act[i] != act_new:
                changed = True
            self._values_act[i] = act_new

        if changed:
            self._parent_universe.channel_changed(self)
        return self

    def to_buffer(self, buf: bytearray):
        byte_order = self._byte_order
        byte_size = self._byte_size

        start = self._buf_start
        for value in self._values_act:
            buf[start: start + byte_size] = value.to_bytes(byte_size, byte_order, signed=False)
            start += byte_size
        return self

    def add_fade(self, values: Collection[Union[int, FadeBase]], duration_ms: int,
                 fade_class: Type[FadeBase] = LinearFade):
        warnings.warn(
            f"{self.set_fade.__name__:s} is deprecated, use {self.set_fade.__name__:s} instead", DeprecationWarning)
        return self.set_fade(values, duration_ms, fade_class)

    # noinspection PyProtectedMember
    def set_fade(self, values: Collection[Union[int, FadeBase]], duration_ms: int,
                 fade_class: Type[FadeBase] = LinearFade):
        """Add and schedule a new fade for the channel

        :param values: Target values for the fade
        :param duration_ms: Duration for the fade in ms
        :param fade_class: What kind of fade
        """
        # check that we passed all values
        if len(values) != self._width:
            raise ValueCountDoesNotMatchChannelWidthError(
                f'Not enough fade values specified, expected {self._width} but got {len(values)}!')

        if self._current_fade is not None:
            self._current_fade.cancel()
            self._current_fade = None

        # calculate how much steps we will be having
        step_time_ms = int(self._parent_node._process_every * 1000)
        duration_ms = max(duration_ms, step_time_ms)
        fade_steps: int = ceil(duration_ms / step_time_ms)

        # build fades
        fades: List[FadeBase] = []
        for i, target in enumerate(values):
            # default is linear
            k = fade_class() if not isinstance(target, FadeBase) else target
            fades.append(k)

            if not 0 <= target <= self._value_max:
                raise ChannelValueOutOfBoundsError(
                    f'Target value out of bounds! 0 <= {target} <= {self._value_max}')

            k.initialize(self._values_raw[i], target, fade_steps)

        # Add to scheduling
        self._current_fade = ChannelBoundFade(self, fades)
        self._parent_node._process_jobs.append(self._current_fade)

        # start fade/refresh task if necessary
        self._parent_node._process_task.start()

        # todo: this on the ChannelBoundFade
        if log.isEnabledFor(LVL_DEBUG):
            log.debug(f'Added fade with {fade_steps} steps:')
            for i, fade in enumerate(fades):
                log.debug(f'CH {self._start + i}: {fade.debug_initialize():s}')
        return self

    def __await__(self):
        if self._current_fade is None:
            return False
        yield from self._current_fade.event.wait().__await__()
        return True

    def __repr__(self):
        return f'<{self.__class__.__name__:s} {self._start:d}/{self._width:d} {self._byte_size * 8:d}bit>'



from . import errors, fades, output_correction
from .__version__ import __version__

# isort: split

from .base import BaseUniverse, Channel

# isort: split

from .impl_artnet import ArtNetNode
from .impl_kinet import KiNetNode
from .impl_sacn import SacnNode



import logging
from time import monotonic
from typing import Dict, Final, Literal

import pyartnet
from pyartnet.errors import ChannelExistsError, ChannelNotFoundError, \
    InvalidUniverseAddressError, OverlappingChannelError

from .output_correction import OutputCorrection

log = logging.getLogger('pyartnet.Universe')


# noinspection PyProtectedMember
class BaseUniverse(OutputCorrection):
    def __init__(self, node: 'pyartnet.base.BaseNode', universe: int = 0):
        super().__init__()

        if not 0 <= universe <= 32767:
            raise InvalidUniverseAddressError()

        self._node: Final = node
        self._universe: Final = universe

        self._data: bytearray = bytearray()
        self._data_size: int = 0
        self._data_changed = True
        self._last_send: float = 0

        self._channels: Dict[str, 'pyartnet.base.Channel'] = {}

    def _apply_output_correction(self):
        for c in self._channels.values():
            c._apply_output_correction()

    def channel_changed(self, channel: 'pyartnet.base.Channel'):
        # update universe buffer
        channel.to_buffer(self._data)

        # signal that this universe has changed
        self._data_changed = True

        # start fade/refresh task if necessary
        # noinspection PyProtectedMember
        self._node._process_task.start()

    def send_data(self):
        self._node._send_universe(self._universe, self._data_size, self._data, self)
        self._last_send = monotonic()
        self._data_changed = False

    def get_channel(self, channel_name: str) -> 'pyartnet.base.Channel':
        """Return a channel by name or raise an exception

        :param channel_name: name of the channel
        """
        if not isinstance(channel_name, str):
            raise TypeError('Channel name must be str')

        try:
            return self._channels[channel_name]
        except KeyError:
            raise ChannelNotFoundError(f'Channel "{channel_name}" not found in the universe!') from None

    def add_channel(self,
                    start: int, width: int,
                    channel_name: str = '',
                    byte_size: int = 1, byte_order: Literal['big', 'little'] = 'little') -> 'pyartnet.base.Channel':
        """Add a new channel to the universe. This will automatically resize the universe accordingly.

        :param start: start position in the universe
        :param width: how many values the channel has
        :param channel_name: name of the channel for requesting it from the universe
        :param byte_size: byte size of a value
        :param byte_order: byte order of a value
        """

        chan = pyartnet.base.Channel(self, start, width, byte_size=byte_size, byte_order=byte_order)

        # build name if not supplied
        if not channel_name:
            channel_name = f'{start:d}/{width:d}'

        # Make sure we don't accidentally overwrite the channel
        if channel_name in self._channels:
            raise ChannelExistsError(f'Channel "{channel_name}" does already exist in the universe!')

        # Make sure channels are not overlapping because they will overwrite each other
        # and this leads to unintended behavior
        for _n, _c in self._channels.items():
            if _c._start > chan._stop or _c._stop < chan._start:
                continue
            for i in range(_c._start, _c._stop + 1):
                if start <= i <= chan._stop:
                    raise OverlappingChannelError(f'New channel {channel_name} is overlapping with channel {_n:s}!')

        self._resize_universe(chan._stop)

        # add channel to universe
        self._channels[channel_name] = chan
        log.debug(f'Added channel "{channel_name}": start: {start:d}, stop: {start + width - 1:d}')

        chan._apply_output_correction()
        return chan

    def _resize_universe(self, min_size: int):

        new_size = max(min_size, 2)
        for c in self._channels.values():
            new_size = max(new_size, c._stop)
        if new_size % 2:
            new_size += 1

        diff = new_size - self._data_size
        if not diff:
            return None

        self._data_size = new_size
        if diff < 0:
            for _ in range(- diff):
                self._data.pop()
        else:
            for _ in range(diff):
                # pad universe data with 0 is it's off
                self._data.append(0)

    # -----------------------------------------------------------
    # emulate container
    def __len__(self):
        return len(self._channels)

    def __getitem__(self, item: str) -> 'pyartnet.base.Channel':
        return self.get_channel(item)



import logging
import socket
from asyncio import sleep
from time import monotonic
from typing import Dict, Final, Generic, List, Optional, Tuple, TypeVar, Union

import pyartnet

from ..errors import DuplicateUniverseError, UniverseNotFoundError
from .background_task import ExceptionIgnoringTask, SimpleBackgroundTask
from .output_correction import OutputCorrection

log = logging.getLogger('pyartnet.ArtNetNode')


TYPE_U = TypeVar('TYPE_U', bound='pyartnet.base.BaseUniverse')


# noinspection PyProtectedMember
class BaseNode(Generic[TYPE_U], OutputCorrection):
    def __init__(self, ip: str, port: int, *,
                 max_fps: int = 25,
                 refresh_every: Union[int, float, None] = 2, start_refresh_task: bool = True,
                 source_address: Optional[Tuple[str, int]] = None):
        super().__init__()

        # Destination
        self._ip: Final = ip
        self._port: Final = port
        self._dst: Final = (self._ip, self._port)

        # socket setup
        self._socket: Final = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP
        self._socket.setblocking(False)  # nonblocking for true asyncio

        # option to set source port/ip
        if source_address is not None:
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._socket.bind(source_address)

        # Name used for the Tasks (e.g. in error msg)
        name: Final = f'{self._ip:s}:{self._port}'

        # refresh task
        self._refresh_every: float = max(0.1, refresh_every)
        self._refresh_task: Final = ExceptionIgnoringTask(self._periodic_refresh_worker, f'Process task {name:s}')
        if start_refresh_task:
            self._refresh_task.start()

        # fade task
        self._process_every: float = 1 / max(1, max_fps)
        self._process_task: Final = SimpleBackgroundTask(self._process_values_task, f'Refresh task {name:s}')
        self._process_jobs: List['pyartnet.base.ChannelBoundFade'] = []

        # packet data
        self._packet_base: Union[bytearray, bytes] = bytearray()
        self._last_send: float = 0

        # containing universes
        self._universes: Tuple[TYPE_U, ...] = ()
        self._universe_map: Dict[int, TYPE_U] = {}

    def _apply_output_correction(self):
        for u in self._universes:
            u._apply_output_correction()

    def _send_universe(self, id: int, byte_size: int, values: bytearray, universe: TYPE_U):
        raise NotImplementedError()

    def _send_data(self, data: Union[bytearray, bytes]) -> int:

        ret = self._socket.sendto(self._packet_base + data, self._dst)

        self._last_send = monotonic()
        return ret

    async def _process_values_task(self):
        # wait a little, so we can schedule multiple tasks/updates, and they all start together
        await sleep(0.01)

        idle_ct = 0
        while idle_ct < 10:
            idle_ct += 1

            # process jobs
            to_remove = []
            for job in self._process_jobs:
                job.process()
                idle_ct = 0

                if job.is_done:
                    to_remove.append(job)

            # send data of universe
            for universe in self._universes:
                if not universe._data_changed:
                    continue
                universe.send_data()
                idle_ct = 0

            if to_remove:
                for job in to_remove:
                    self._process_jobs.remove(job)
                    job.fade_complete()

            await sleep(self._process_every)

    def start_refresh(self):
        """Manually start the refresh task (if not already running)"""
        self._refresh_task.start()

    def stop_refresh(self):
        """Manually stop the refresh task"""
        self._refresh_task.cancel()

    async def _periodic_refresh_worker(self):
        while True:
            # sync the refresh messages
            next_refresh = monotonic()
            for u in self._universes:
                next_refresh = min(next_refresh, u._last_send)

            diff = monotonic() - next_refresh
            if diff < self._refresh_every:
                await sleep(diff)
                continue

            for u in self._universes:
                u.send_data()

    def get_universe(self, nr: int) -> TYPE_U:
        """Get universe by number

        :param nr: universe nr
        :return: The universe
        """
        if not isinstance(nr, int) or not nr >= 0:
            raise ValueError('BaseUniverse must be an int >= 0!')
        nr = int(nr)

        try:
            return self._universe_map[nr]
        except KeyError:
            raise UniverseNotFoundError(f'BaseUniverse {nr:d} not found!') from None

    def add_universe(self, nr: int = 0) -> TYPE_U:
        """Creates a new universe and adds it to the parent node

        :param nr: universe nr
        :return: The universe
        """
        if not isinstance(nr, int) or not nr >= 0:
            raise ValueError('BaseUniverse must be an int >= 0!')
        nr = int(nr)

        if nr in self._universe_map:
            raise DuplicateUniverseError(f'BaseUniverse {nr:d} does already exist!')

        # add to data
        self._universe_map[nr] = universe = self._create_universe(nr)
        self._universes = tuple(u for _, u in sorted(self._universe_map.items()))   # ascending

        return universe

    def _create_universe(self, nr: int) -> TYPE_U:
        raise NotImplementedError()

    def __await__(self):
        while self._process_jobs:
            for job in self._process_jobs:
                yield from job.channel.__await__()

    def __getitem__(self, nr: int) -> TYPE_U:
        return self.get_universe(nr)

    def __len__(self):
        return len(self._universes)




class FadeBase:

    def __init__(self):
        self.is_done = False

    def initialize(self, current: int, target: int, steps: int):
        raise NotImplementedError()

    def debug_initialize(self) -> str:
        """return debug string of the calculated values in initialize fade"""
        return ""

    def calc_next_value(self) -> float:
        raise NotImplementedError()



class PyArtNetError(Exception):
    pass


# -----------------------------------------------------------------------------
# Implementation specific Errors
# -----------------------------------------------------------------------------
class InvalidCidError(PyArtNetError):
    pass


# -----------------------------------------------------------------------------
# Universe Errors
# -----------------------------------------------------------------------------
class InvalidUniverseAddressError(PyArtNetError):
    pass


class DuplicateUniverseError(PyArtNetError):
    pass


class UniverseNotFoundError(PyArtNetError):
    pass


# -----------------------------------------------------------------------------
# Channel Errors
# -----------------------------------------------------------------------------
class ChannelExistsError(PyArtNetError):
    pass


class ChannelNotFoundError(PyArtNetError):
    pass


class OverlappingChannelError(PyArtNetError):
    pass


class ChannelOutOfUniverseError(PyArtNetError):
    pass


class ChannelWidthError(PyArtNetError):
    pass


class ChannelValueOutOfBoundsError(PyArtNetError):
    pass


class ValueCountDoesNotMatchChannelWidthError(PyArtNetError):
    pass



import logging
from typing import Final, Optional, Tuple, Union

import pyartnet
from pyartnet.base import BaseNode
from pyartnet.base.seq_counter import SequenceCounter
from pyartnet.errors import InvalidUniverseAddressError

# -----------------------------------------------------------------------------
# Documentation for ArtNet Protocol:
# https://artisticlicence.com/support-and-resources/art-net-4/
# -----------------------------------------------------------------------------

log = logging.getLogger('pyartnet.ArtNetNode')


class ArtNetNode(BaseNode['pyartnet.impl_artnet.ArtNetUniverse']):
    def __init__(self, ip: str, port: int, *,
                 max_fps: int = 25,
                 refresh_every: Union[int, float, None] = 2, start_refresh_task: bool = True,
                 source_address: Optional[Tuple[str, int]] = None,

                 # ArtNet specific fields
                 sequence_counter: bool = True
                 ):
        super().__init__(ip=ip, port=port,
                         max_fps=max_fps,
                         refresh_every=refresh_every, start_refresh_task=start_refresh_task,
                         source_address=source_address)

        # ArtNet specific fields
        self._sequence_ctr: Final = SequenceCounter(1) if sequence_counter else SequenceCounter(0, 0)

        # build base packet
        packet = bytearray()
        packet.extend(map(ord, "Art-Net"))
        packet.append(0x00)          # Null terminate Art-Net
        packet.extend([0x00, 0x50])  # Opcode ArtDMX 0x5000 (Little endian)
        packet.extend([0x00, 0x0e])  # Protocol version 14
        self._packet_base = bytes(packet)

    def _send_universe(self, id: int, byte_size: int, values: bytearray,
                       universe: 'pyartnet.impl_artnet.ArtNetUniverse'):

        # pre allocate the bytearray
        _size = 6 + byte_size
        packet = bytearray(_size)

        packet[0] = self._sequence_ctr.value                    # 1 | Sequence,
        packet[1] = 0x00                                        # 1 | Physical input port (not used)
        packet[2:4] = id.to_bytes(2, byteorder='little')        # 2 | Universe

        packet[4:6] = byte_size.to_bytes(2, 'big')              # 2       | Number of channels Big Endian
        packet[6: _size] = values                               # 0 - 512 | Channel values

        self._send_data(packet)

        # log complete packet
        if log.isEnabledFor(logging.DEBUG):
            self.__log_artnet_frame(self._packet_base + packet)

    def _create_universe(self, nr: int) -> 'pyartnet.impl_artnet.ArtNetUniverse':
        if nr >= 32_768:
            raise InvalidUniverseAddressError()
        return pyartnet.impl_artnet.ArtNetUniverse(self, nr)

    def __log_artnet_frame(self, p: Union[bytearray, bytes]):
        """Log Artnet Frame"""
        assert isinstance(p, (bytearray, bytes))

        # runs the first time
        if not hasattr(self, '_log_ctr'):
            self._log_ctr = -1
            self._log_show = [False for k in range(103)]

        self._log_ctr += 1
        if self._log_ctr >= 10:
            self._log_ctr = 0
        show_description: bool = self._log_ctr == 0

        host_fmt = ' ' * (36 + len(self._ip))
        out_desc = '{:s} {:2s} {:2s} {:4s} {:4s}'.format(host_fmt, 'Sq', '', 'Univ', ' Len')

        _max_channel = p[16] << 8 | p[17]
        pre = bytearray(p[:12]).hex().upper()
        out = f'Packet to {self._ip:s}: {pre} {p[12]:02x} {p[13]:02x} {p[13]:02x}{p[14]:02x} {_max_channel:04x}'

        # check what to print
        for k in range(_max_channel):
            if p[18 + k]:
                # once we change something print channel index
                if self._log_show[k // 5] is False:
                    show_description = True
                self._log_show[k // 5] = True

        for k in range(0, _max_channel, 5):

            # if there was never anything active do not print, but print the last block
            if not self._log_show[k // 5] and not k + 5 > _max_channel:
                # do not print multiple dots
                if out.endswith('...'):
                    continue

                out_desc += '  - '
                out += ' ...'
                continue

            # format block of channels
            _block_vals = []
            _block_desc = []
            for i in range(5):
                if k + i < _max_channel:
                    if show_description:
                        _block_desc.append(f'{k + i + 1:<3d}')
                    _block_vals.append(f'{p[18 + k + i]:03d}')

            # separator
            if out.endswith('...'):
                out_desc += ' '
                out += ' '
            else:
                out_desc += '   '
                out += '   '

            out += ' '.join(_block_vals)
            if show_description:
                out_desc += ' '.join(_block_desc)

        if show_description:
            log.debug(out_desc)
        log.debug(out)



from time import monotonic

import pytest

from pyartnet.base import BaseUniverse
from pyartnet.base.channel import Channel
from pyartnet.errors import DuplicateUniverseError
from tests.conftest import STEP_MS, TestingNode


def test_universe_add_get(node: TestingNode):
    for i in (1.3, -1):
        with pytest.raises(ValueError, match='BaseUniverse must be an int >= 0!'):
            node.add_universe(i)

        with pytest.raises(ValueError, match='BaseUniverse must be an int >= 0!'):
            node.get_universe(i)

    u = node.add_universe()
    assert len(node) == 1
    assert node.get_universe(0) is u
    assert node[0] is u

    # Duplicate
    with pytest.raises(DuplicateUniverseError, match='BaseUniverse 0 does already exist!'):
        node.add_universe()

    # Check that the nodes are ascending
    node.add_universe(50)
    node.add_universe(3)

    assert len(node) == 3
    assert node._universes == (
        node.get_universe(0), node.get_universe(3), node.get_universe(50)
    )


async def test_fade_await(node: TestingNode, universe: BaseUniverse, caplog):
    async def check_no_wait_time_when_no_fade():
        start = monotonic()
        for _ in range(1000):
            assert not await node
        assert monotonic() - start < 0.001

    async def check_wait_time_when_fade(steps: int):
        start = monotonic()
        await node
        assert monotonic() - start >= ((steps - 1) * STEP_MS) / 1000

    caplog.set_level(0)

    channel = Channel(universe, 1, 1)

    await check_no_wait_time_when_no_fade()

    channel.set_fade([2], 2 * STEP_MS)
    assert channel.get_values() == [0]

    assert list(caplog.messages) == [
        'Added fade with 2 steps:',
        'CH 1: 000 -> 002 | step:  +1.0'
    ]

    assert channel._current_fade is not None
    await check_wait_time_when_fade(2)
    assert channel._current_fade is None
    assert channel.get_values() == [2]
    assert node.data == ['01', '02']

    await check_no_wait_time_when_no_fade()

    channel.set_fade([10], 2 * STEP_MS)

    assert channel._current_fade is not None
    await check_wait_time_when_fade(2)
    assert channel._current_fade is None
    assert node.data == ['01', '02', '06', '0a']

    await check_no_wait_time_when_no_fade()
    await node.wait_for_task_finish()



from unittest.mock import Mock

from pyartnet.base import Channel
from pyartnet.base.channel_fade import ChannelBoundFade


def test_repr():
    universe = Mock()
    universe.output_correction = None

    a = Channel(universe, 1, 2)

    a = ChannelBoundFade(a, [])
    assert repr(a) == '<ChannelBoundFade channel=1/2, is_done=False>'

    a.is_done = True
    assert repr(a) == '<ChannelBoundFade channel=1/2, is_done=True>'

    a = ChannelBoundFade(a, [])
    a.channel = None
    assert repr(a) == '<ChannelBoundFade channel=None, is_done=False>'



import logging
from logging import DEBUG as LVL_DEBUG
from struct import pack as s_pack
from typing import Optional, Tuple, Union

import pyartnet
from pyartnet.base import BaseNode
from pyartnet.errors import InvalidUniverseAddressError

# -----------------------------------------------------------------------------
# Documentation for KiNet Protocol:
# todo: find links
# -----------------------------------------------------------------------------

log = logging.getLogger('pyartnet.KiNetNode')


class KiNetNode(BaseNode['pyartnet.impl_kinet.KiNetUniverse']):
    def __init__(self, ip: str, port: int, *,
                 max_fps: int = 25,
                 refresh_every: Union[int, float, None] = 2, start_refresh_task: bool = True,
                 source_address: Optional[Tuple[str, int]] = None):
        super().__init__(ip=ip, port=port,
                         max_fps=max_fps,
                         refresh_every=refresh_every, start_refresh_task=start_refresh_task,
                         source_address=source_address)

        # build base packet
        packet = bytearray()
        packet.extend(s_pack(">IHH", 0x0401DC4A, 0x0100, 0x0101))   # Magic, version, type
        packet.extend(s_pack(">IBBHI", 0, 0, 0, 0, 0xFFFFFFFF))     # sequence, port, padding, flags, timer
        self._packet_base = bytes(packet)

    def _send_universe(self, id: int, byte_size: int, values: bytearray, universe: 'pyartnet.impl_kinet.KiNetUniverse'):
        packet = bytearray()
        packet.append(byte_size)
        packet.extend(values)

        self._send_data(packet)

        if log.isEnabledFor(LVL_DEBUG):
            # log complete packet
            log.debug(f"Sending KiNet frame to {self._ip}:{self._port}: {(self._packet_base + packet).hex()}")

    def _create_universe(self, nr: int) -> 'pyartnet.impl_kinet.KiNetUniverse':
        if nr >= 32_768:
            raise InvalidUniverseAddressError()
        return pyartnet.impl_kinet.KiNetUniverse(self, nr)



# flake8: noqa: E262
import logging
from logging import DEBUG as LVL_DEBUG
from typing import Final, Optional, Tuple, Union
from uuid import uuid4

import pyartnet.impl_sacn.universe
from pyartnet.base import BaseNode
from pyartnet.errors import InvalidCidError, InvalidUniverseAddressError

# -----------------------------------------------------------------------------
# Documentation for E1.31 Protocol:
# https://tsp.esta.org/tsp/documents/published_docs.php
# -----------------------------------------------------------------------------

log = logging.getLogger('pyartnet.SacnNode')


# Package constant
ACN_PACKET_IDENTIFIER: Final = (0x41, 0x53, 0x43, 0x2d, 0x45, 0x31, 0x2e, 0x31, 0x37, 0x00, 0x00, 0x00)

# Field constants
VECTOR_ROOT_E131_DATA: Final = b'\x00\x00\x00\x04'
VECTOR_E131_DATA_PACKET: Final = b'\x00\x00\x00\x02'
VECTOR_DMP_SET_PROPERTY: Final = 0x02


class SacnNode(BaseNode['pyartnet.impl_sacn.SacnUniverse']):
    def __init__(self, ip: str, port: int, *,
                 max_fps: int = 25,
                 refresh_every: Union[int, float, None] = 2, start_refresh_task: bool = True,
                 source_address: Optional[Tuple[str, int]] = None,

                 # sACN E1.31 specific fields
                 cid: Optional[bytes] = None, source_name: Optional[str] = None
                 ):
        super().__init__(ip=ip, port=port,
                         max_fps=max_fps,
                         refresh_every=refresh_every, start_refresh_task=start_refresh_task,
                         source_address=source_address)

        # CID Field
        if cid is not None:
            if not isinstance(cid, bytes) or len(cid) != 16:
                raise InvalidCidError('CID must be 16bytes!')
        else:
            cid = uuid4().bytes

        # Source field
        if source_name is None:
            source_name = 'PyArtNet'
        source_name_byte = source_name.encode('utf-8').ljust(64, b'\x00')
        if len(source_name_byte) > 64:
            raise ValueError('Source name too long!')

        # build base packet
        packet = bytearray()

        # Root layer
        packet.extend(b'\x00\x10')              # |  2 | Preamble Size
        packet.extend(b'\x00\x00')              # |  2 | Post-amble Size
        packet.extend(ACN_PACKET_IDENTIFIER)    # | 12 | Packet Identifier
        packet.extend([0x72, 0x57])             # |  2 | Flags, Length
        packet.extend(VECTOR_ROOT_E131_DATA)    # |  4 | Vector
        packet.extend(cid)                      # | 16 | CID, a unique identifier

        # Framing layer Part 1
        packet.extend([0x72, 0x57])                 # |  2 | Flags and length
        packet.extend(VECTOR_E131_DATA_PACKET)      # |  4 | Vector
        packet.extend(source_name_byte)             # | 64 |Source Name
        packet.append(100)                          # |  1 |Priority
        packet.extend(int(50).to_bytes(2, 'big'))   # |  2 | Synchronization universe

        self._packet_base: bytearray = packet


    def _send_universe(self, id: int, byte_size: int, values: bytearray,
                       universe: 'pyartnet.impl_sacn.universe.SacnUniverse'):
        packet = bytearray()

        # DMX Start Code is not included in the byte size from the universe
        prop_count = byte_size + 1

        # Framing layer Part 2
        packet.append(universe._sequence_ctr.value)             # | 1 | Sequence,
        packet.append(0x00)                                     # | 1 | Options
        packet.extend(id.to_bytes(2, byteorder='big'))          # | 2 | BaseUniverse Number

        # DMP Layer
        dmp_length = ((10 + prop_count) | 0x7000).to_bytes(2, 'big')
        packet.extend(dmp_length)               # | 2 | Flags and length
        packet.append(VECTOR_DMP_SET_PROPERTY)  # | 1 | Vector
        packet.append(0xA1)                     # | 1 | Address Type & Data Type
        packet.extend(b'\x00\x00')              # | 2 | First Property Address
        packet.extend(b'\x00\x01')              # | 2 | Address Increment

        packet.extend(prop_count.to_bytes(2, 'big'))    # |     2 | Property Value Count
        packet.append(0x00)                             # |     1 | Property Values - DMX Start Code
        packet.extend(values)                           # | 0-512 | Property Values - DMX Data

        # Update length for base packet
        base_packet = self._packet_base
        base_packet[16:18] = ((109 + prop_count) | 0x7000).to_bytes(2, 'big')   # root layer
        base_packet[38:40] = (( 87 + prop_count) | 0x7000).to_bytes(2, 'big')   # framing layer

        self._send_data(packet)

        if log.isEnabledFor(LVL_DEBUG):
            # log complete packet
            log.debug(f"Sending sACN frame to {self._ip}:{self._port}: {(base_packet + packet).hex()}")

    def _create_universe(self, nr: int) -> 'pyartnet.impl_sacn.SacnUniverse':
        # 6.2.7 E1.31 Data Packet: Universe
        if not 1 <= nr < 63_999:
            raise InvalidUniverseAddressError()
        return pyartnet.impl_sacn.SacnUniverse(self, nr)



import logging
from asyncio import create_task, sleep, Task
from time import monotonic
from traceback import format_exc
from typing import Any, Callable, Coroutine, Final, Optional, Set

log = logging.getLogger('pyartnet.Task')


def log_exception(e: Exception, name: str):
    log.error(f'Error in worker for {name:s}:')
    for line in format_exc().splitlines():
        log.error(line)


_BACKGROUND_TASKS: Set[Task] = set()

# use variables, so it's easy to e.g. implement thread safe scheduling
CREATE_TASK = create_task
EXCEPTION_HANDLER: Callable[[Exception, str], Any] = log_exception


class SimpleBackgroundTask:

    def __init__(self, coro: Callable[[], Coroutine], name: str):
        self.coro: Final = coro
        self.name: Final = name
        self.task: Optional[Task] = None

    def start(self):
        if self.task is not None:
            return None

        self.task = task = CREATE_TASK(self.coro_wrap(), name=self.name)
        _BACKGROUND_TASKS.add(task)
        task.add_done_callback(_BACKGROUND_TASKS.discard)

    def cancel(self):
        if self.task is None:
            return None

        self.task.cancel()
        self.task = None

    async def coro_wrap(self):
        log.debug(f'Started {self.name}')
        task = self.task
        assert task is not None

        try:
            await self.coro()
        except Exception as e:
            EXCEPTION_HANDLER(e, self.name)
        finally:
            if self.task is task:
                self.task = None
            log.debug(f'Stopped {self.name}')


class ExceptionIgnoringTask(SimpleBackgroundTask):
    async def coro_wrap(self):
        log.debug(f'Started {self.name}')
        task = self.task
        assert task is not None

        wait = 0

        try:
            while True:
                await sleep(wait)
                start = monotonic()
                try:
                    await self.coro()
                except Exception as e:
                    EXCEPTION_HANDLER(e, self.name)

                    # simple sleep logic with an increasing timeout
                    time_to_exception = monotonic() - start
                    if time_to_exception < 16 or time_to_exception < wait:
                        wait = max(2, wait * 2)
                    else:
                        wait = 0

                    log.debug(f'Retry in {wait:d} seconds')
        finally:
            if self.task is task:
                self.task = None
            log.debug(f'Stopped {self.name}')



from typing import Final


class SequenceCounter:
    __slots__ = ('_ctr', '_start', '_upper')

    def __init__(self, start: int = 0, upper: int = 255):
        self._ctr: int = start
        assert start <= upper
        self._start: Final = start
        self._upper: Final = upper

    @property
    def value(self) -> int:
        ret = self._ctr
        self._ctr += 1
        if self._ctr > self._upper:
            self._ctr = self._start
        return ret

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self._ctr:d}>'



import pytest

from pyartnet import errors
from pyartnet.base import BaseUniverse
from pyartnet.errors import ChannelNotFoundError


def test_exceptions(universe: BaseUniverse):
    universe.add_channel(1, 1)

    with pytest.raises(errors.ChannelExistsError) as e:
        universe.add_channel(1, 1)
    assert str(e.value) == 'Channel "1/1" does already exist in the universe!'

    # Channels
    with pytest.raises(errors.ChannelNotFoundError) as e:
        universe.get_channel('2')
    assert str(e.value) == 'Channel "2" not found in the universe!'
    with pytest.raises(errors.ChannelNotFoundError) as e:
        _ = universe['2']
    assert str(e.value) == 'Channel "2" not found in the universe!'

    # Overlapping channels
    universe.add_channel(10, 3)

    with pytest.raises(errors.OverlappingChannelError) as e:
        universe.add_channel(1, 3)
    assert str(e.value) == 'New channel 1/3 is overlapping with channel 1/1!'

    with pytest.raises(errors.OverlappingChannelError):
        universe.add_channel(9, 2)
    with pytest.raises(errors.OverlappingChannelError):
        universe.add_channel(12, 1)
    with pytest.raises(errors.OverlappingChannelError):
        universe.add_channel(8, 20)


def test_universe_resize(universe: BaseUniverse):
    assert universe._data_size == 0
    assert universe._data == b''

    universe.add_channel(1, 1)
    assert universe._data_size == 2
    assert universe._data == b'\x00\x00'

    universe.add_channel(6, 1)
    assert universe._data_size == 6
    assert universe._data == b'\x00\x00\x00\x00\x00\x00'

    universe._channels.popitem()
    universe.add_channel(2, 1)
    assert universe._data_size == 2
    assert universe._data == b'\x00\x00'

    universe.add_channel(3, 1)
    assert universe._data_size == 4
    assert universe._data == b'\x00\x00\x00\x00'


def test_access(universe: BaseUniverse):

    with pytest.raises(ChannelNotFoundError) as e:
        universe.get_channel('1')
    assert str(e.value) == 'Channel "1" not found in the universe!'

    with pytest.raises(ChannelNotFoundError) as e:
        universe.get_channel('1/1')
    assert str(e.value) == 'Channel "1/1" not found in the universe!'

    c = universe.add_channel(1, 1)
    assert len(universe) == 1
    assert universe.get_channel('1/1') is c
    assert universe['1/1'] is c

    c = universe.add_channel(2, 1)
    assert len(universe) == 2
    assert universe.get_channel('2/1') is c
    assert universe['2/1'] is c



def linear(val: float, max_val: int = 0xFF) -> float:
    """linear output correction"""
    return val


def quadratic(val: float, max_val: int = 0xFF) -> float:
    """Quadratic output correction"""
    return (val ** 2) / max_val


def cubic(val: float, max_val: int = 0xFF) -> float:
    """Cubic output correction"""
    return (val ** 3) / (max_val ** 2)


def quadruple(val: float, max_val: int = 0xFF) -> float:
    """Quadruple output correction"""
    return (val ** 4) / (max_val ** 3)



from typing import Callable, Optional


class OutputCorrection:
    def __init__(self):
        super().__init__()
        self._correction_output: Optional[Callable[[float, int], float]] = None

    def set_output_correction(self, func: Optional[Callable[[float, int], float]]) -> None:
        """Set the output correction function.

        :param func: None to disable output correction or the function which will be used to transform the values
        """
        self._correction_output = func
        self._apply_output_correction()
        return None

    def _apply_output_correction(self) -> None:
        raise NotImplementedError()



from .fade_base import FadeBase


class LinearFade(FadeBase):

    def __init__(self):
        super().__init__()
        self.target: int = 0            # Target Value
        self.current: float = 0.0       # Current Value
        self.factor: float = 1.0

    def debug_initialize(self) -> str:
        return f"{self.current:03.0f} -> {self.target:03d} | step: {self.factor:+5.1f}"

    def initialize(self, start: int, target: int, steps: int):
        self.current = start
        self.target = target
        self.factor = (self.target - start) / steps

    def calc_next_value(self) -> float:
        self.current += self.factor

        # is_done status
        curr = round(self.current)
        if self.factor <= 0:
            if curr <= self.target:
                self.is_done = True
        else:
            if curr >= self.target:
                self.is_done = True

        return self.current



from pyartnet.base.seq_counter import SequenceCounter


def test_seq():
    s = SequenceCounter()
    assert s.value == 0
    assert s.value == 1
    assert s.value == 2

    s._ctr = 254
    assert s.value == 254
    assert s.value == 255
    assert s.value == 0
    assert s.value == 1


def test_seq_artnet():
    s = SequenceCounter(1)
    assert s.value == 1
    assert s.value == 2

    s._ctr = 254
    assert s.value == 254
    assert s.value == 255
    assert s.value == 1


def test_seq_const():
    s = SequenceCounter(0, 0)
    assert s.value == 0
    assert s.value == 0
    assert s.value == 0


def test_repr():
    s = SequenceCounter()
    assert repr(s) == '<SequenceCounter 0>'
    assert repr(s) == '<SequenceCounter 0>'
    assert s.value == 0
    assert repr(s) == '<SequenceCounter 1>'



import logging
from asyncio import Event
from typing import Final, Iterable, List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    import pyartnet


log = logging.getLogger('pyartnet.DmxChannel')


# noinspection PyProtectedMember
class ChannelBoundFade:
    def __init__(self, channel: 'pyartnet.base.Channel', fades: Iterable['pyartnet.fades.FadeBase']):
        super().__init__()
        self.channel: 'pyartnet.base.Channel' = channel

        self.fades: Tuple['pyartnet.fades.FadeBase', ...] = tuple(fades)
        self.values: List[float] = [0 for _ in fades]

        self.is_done = False
        self.event: Final = Event()

    def process(self):
        finished = True
        for i, fade in enumerate(self.fades):
            if fade.is_done:
                continue

            self.values[i] = fade.calc_next_value()

            if not fade.is_done:
                finished = False

        self.is_done = finished
        self.channel.set_values(self.values)

    def cancel(self):
        # remove fade from channel
        c = self.channel
        self.channel = None  # type: ignore[assignment]
        c._current_fade = None

        self.event.set()

        # remove from parent node
        c._parent_node._process_jobs.remove(self)

    def fade_complete(self):
        # remove fade from channel
        c = self.channel
        self.channel = None  # type: ignore[assignment]
        c._current_fade = None

        self.event.set()

        if c.callback_fade_finished is not None:
            c.callback_fade_finished(c)

    def __repr__(self):
        # Channel part
        if self.channel is not None:
            channel_part = f'channel={self.channel._start:d}/{self.channel._width:d}'
        else:
            channel_part = 'channel=None'

        return f'<{self.__class__.__name__:s} {channel_part}, is_done={self.is_done}>'



from typing import Final

import pyartnet
from pyartnet.base import BaseUniverse
from pyartnet.base.seq_counter import SequenceCounter


class SacnUniverse(BaseUniverse):

    def __init__(self, node: 'pyartnet.impl_sacn.SacnNode', universe: int = 0):
        super().__init__(node, universe)

        # sACN has the sequence counter on the universe
        self._sequence_ctr: Final = SequenceCounter()
